@@ -31,38 +31,32 @@
 class Soprano::Node::NodeData : public QSharedData
 {
 public:
-    NodeData( Type type_ = EmptyNode )
-        : type( type_ ) {}
-
+    NodeData() {}
     virtual ~NodeData() {}
 
-    Type type;
-
-    virtual QString toString() const {
-        return QString();
-    }
-
-    virtual QString toN3() const {
-        return QString();
-    }
+    virtual Type type() const = 0;
+    virtual QString toString() const = 0;
+    virtual QString toN3() const = 0;
 };
 
 class Soprano::Node::ResourceNodeData : public NodeData
 {
 public:
     ResourceNodeData( const QUrl& uri_ = QUrl() )
-        : NodeData( ResourceNode ),
+        : NodeData(),
           uri( uri_ ){
     }
 
     QUrl uri;
 
+    Type type() const { return ResourceNode; }
+
     QString toString() const {
         return uri.toString();
     }
 
     QString toN3() const {
-        return '<' + QString::fromAscii( uri.toEncoded() ) + '>';
+        return Node::resourceToN3( uri );
     }
 };
 
@@ -70,45 +64,42 @@ class Soprano::Node::BNodeData : public NodeData
 {
 public:
     BNodeData( const QString& id = QString() )
-        : NodeData( BlankNode ),
+        : NodeData(),
           identifier( id ) {
     }
 
     QString identifier;
 
+    Type type() const { return BlankNode; }
+
     QString toString() const {
         return identifier;
     }
 
     QString toN3() const {
-        return "_:" + identifier;
+        return Node::blankToN3( identifier );
     }
 };
 
 class Soprano::Node::LiteralNodeData : public NodeData
 {
 public:
-    LiteralNodeData( const LiteralValue& val = LiteralValue(), const QString& lang = QString() )
-        : NodeData( LiteralNode ),
-          value( val ),
-          language( lang ) {
+    LiteralNodeData( const LiteralValue& val = LiteralValue() )
+        : NodeData(),
+          value( val )
+    {
     }
 
     LiteralValue value;
-    QString language;
+
+    Type type() const { return LiteralNode; }
 
     QString toString() const {
         return value.toString();
     }
 
     QString toN3() const {
-        if( !language.isEmpty() ) {
-            return '\"' + toString().replace( '\"', "\\\"" ) + "\"@" + language;
-        }
-        else {
-            return QString( "\"%1\"^^<%2>" )
-                .arg( toString().replace( '\"', "\\\"" ), QString::fromAscii( value.dataTypeUri().toEncoded() ) );
-        }
+        return Node::literalToN3( value );
     }
 };
 
@@ -143,10 +134,26 @@ Soprano::Node::Node( const QString &id )
     }
 }
 
+Soprano::Node::Node( const LiteralValue& value )
+{
+    if ( value.isValid() ) {
+        d = new LiteralNodeData( value );
+    }
+    else {
+        d = 0;
+    }
+}
+
 Soprano::Node::Node( const LiteralValue& value, const QString& lang )
 {
     if ( value.isValid() ) {
-        d = new LiteralNodeData( value, lang );
+        LiteralValue lit;
+        if ( lang.isEmpty() ) {
+            lit = value;
+        } else {
+            lit = LiteralValue::createPlainLiteral( value.toString(), lang );
+        }
+        d = new LiteralNodeData( lit );
     }
     else {
         d = 0;
@@ -159,7 +166,7 @@ Soprano::Node::~Node()
 
 bool Soprano::Node::isEmpty() const
 {
-    return ( !d || d->type == Soprano::Node::EmptyNode );
+    return ( !d || d->type() == Soprano::Node::EmptyNode );
 }
 
 bool Soprano::Node::isValid() const
@@ -169,22 +176,22 @@ bool Soprano::Node::isValid() const
 
 bool Soprano::Node::isLiteral() const
 {
-    return ( d && d->type == Soprano::Node::LiteralNode );
+    return ( d && d->type() == Soprano::Node::LiteralNode );
 }
 
 bool Soprano::Node::isResource() const
 {
-    return ( d && d->type == Soprano::Node::ResourceNode );
+    return ( d && d->type() == Soprano::Node::ResourceNode );
 }
 
 bool Soprano::Node::isBlank() const
 {
-    return ( d && d->type == Soprano::Node::BlankNode );
+    return ( d && d->type() == Soprano::Node::BlankNode );
 }
 
 Soprano::Node::Type Soprano::Node::type() const
 {
-    return d ? d->type : EmptyNode;
+    return d ? d->type() : EmptyNode;
 }
 
 QUrl Soprano::Node::uri() const
@@ -231,12 +238,7 @@ QUrl Soprano::Node::dataType() const
 
 QString Soprano::Node::language() const
 {
-    if ( isLiteral() ) {
-        return static_cast<const LiteralNodeData*>( d.constData() )->language;
-    }
-    else {
-        return QString();
-    }
+    return literal().language().toString();
 }
 
 QString Soprano::Node::toString() const
@@ -283,19 +285,17 @@ bool Soprano::Node::operator==( const Node& other ) const
         return false;
     }
     else if ( type() != EmptyNode ) {
-        if ( d->type == ResourceNode ) {
+        if ( d->type() == ResourceNode ) {
             return(  static_cast<const ResourceNodeData*>( d.constData() )->uri ==
                      static_cast<const ResourceNodeData*>( other.d.constData() )->uri );
         }
-        else if ( d->type == BlankNode ) {
+        else if ( d->type() == BlankNode ) {
             return( static_cast<const BNodeData*>( d.constData() )->identifier ==
                     static_cast<const BNodeData*>( other.d.constData() )->identifier );
         }
-        else if ( d->type == LiteralNode ) {
+        else if ( d->type() == LiteralNode ) {
             return ( static_cast<const LiteralNodeData*>( d.constData() )->value ==
-                     static_cast<const LiteralNodeData*>( other.d.constData() )->value &&
-                     static_cast<const LiteralNodeData*>( d.constData() )->language ==
-                     static_cast<const LiteralNodeData*>( other.d.constData() )->language );
+                     static_cast<const LiteralNodeData*>( other.d.constData() )->value );
         }
     }
 
@@ -319,9 +319,7 @@ bool Soprano::Node::operator!=( const Node& other ) const
     }
     else if ( type() == LiteralNode ) {
         return ( static_cast<const LiteralNodeData*>( d.constData() )->value !=
-                 static_cast<const LiteralNodeData*>( other.d.constData() )->value ||
-                 static_cast<const LiteralNodeData*>( d.constData() )->language !=
-                 static_cast<const LiteralNodeData*>( other.d.constData() )->language );
+                 static_cast<const LiteralNodeData*>( other.d.constData() )->value );
     }
     else {
         // empty nodes are always equal
@@ -341,7 +339,8 @@ bool Soprano::Node::operator==( const LiteralValue& other ) const
 
 bool Soprano::Node::matches( const Node& other ) const
 {
-    if ( isEmpty() || other.isEmpty() ) {
+    // other is a template
+    if ( other.isEmpty() ) {
         return true;
     }
     else {
@@ -367,10 +366,61 @@ Soprano::Node Soprano::Node::createBlankNode( const QString& id )
     return Node( id );
 }
 
+Soprano::Node Soprano::Node::createLiteralNode( const LiteralValue& val )
+{
+    return Node( val );
+}
 
 Soprano::Node Soprano::Node::createLiteralNode( const LiteralValue& val, const QString& language )
 {
-    return Node( val, language );
+    return Node( LiteralValue::createPlainLiteral( val.toString(), language ) );
+}
+
+
+QString Soprano::Node::resourceToN3( const QUrl& uri )
+{
+    return '<' + QString::fromAscii( uri.toEncoded() ) + '>';
+}
+
+
+QString Soprano::Node::blankToN3( const QString& blank )
+{
+    if ( blank.isEmpty() )
+        return blank;
+    else
+        return "_:" + blank;
+}
+
+
+QString Soprano::Node::literalToN3( const LiteralValue& literal )
+{
+    //
+    // Escape control chars:  \t \b \n \r \f \\ \" \'
+    //
+    QString s = literal.toString();
+
+    // FIXME: this can be done faster by running through the string only once.
+    s.replace( '\\', "\\\\" );
+    s.replace( '\"', "\\\"" );
+    s.replace( '\'', "\\\'" );
+    s.replace( '\n', "\\n" );
+    s.replace( '\r', "\\r" );
+    s.replace( '\b', "\\b" );
+    s.replace( '\t', "\\t" );
+    s.replace( '\f', "\\f" );
+
+    if( literal.isPlain() ) {
+        if ( literal.language().isEmpty() ) {
+            return '\"' + s + '\"';
+        }
+        else {
+            return '\"' + s + "\"@" + literal.language().toString();
+        }
+    }
+    else {
+        return QString( "\"%1\"^^<%2>" )
+            .arg( s, QString::fromAscii( literal.dataTypeUri().toEncoded() ) );
+    }
 }
 
 
@@ -404,10 +454,32 @@ QTextStream& operator<<( QTextStream& s, const Soprano::Node& n )
 
 uint Soprano::qHash( const Soprano::Node& node )
 {
-    if( node.isResource() )
-        return qHash( node.uri() );
-    else
-        return qHash( node.toString() );
+    uint hashVal;
+    switch ( node.type() ) {
+    case Soprano::Node::EmptyNode:
+        hashVal = 0;
+        break;
+    case Soprano::Node::ResourceNode:
+        hashVal = qHash( node.uri() );
+        break;
+    case Soprano::Node::LiteralNode:
+        hashVal = qHash( node.literal() );
+        break;
+    case Soprano::Node::BlankNode:
+        hashVal = qHash( node.identifier() );
+        break;
+    default:
+        // Should never get here
+        hashVal = 0;
+        break;
+    }
+
+    // Rotate bits so as to further distinguish node types with the
+    // same string
+    uint typeInt( ((uint)node.type()) & 0x1F );
+    hashVal = (hashVal << typeInt) | (hashVal >> (0x20 - typeInt));
+
+    return hashVal;
 }
 
 
